#
#    (C) Quantum Computing Inc., 2020.
#
#    THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE of the Copyright Holder. 
#    The contents of this file may not be disclosed to third parties, copied 
#    or duplicated in any form, in whole or in part, without the prior written
#    permission of the Copyright Holder.
#
from typing import Union
import requests
import sys
import os
from typing import Any
import yaml
import json
import networkx as nx
import numpy as np
import scipy.sparse as sp
import time
from requests_toolbelt.multipart import encoder

from .encoders import encode_data, encode_sparse_array, encode_ndarray
from .decoders import decodeJson
from .utils import TicToc


def get_client_kwargs(**kwargs):
    """
    From kwargs extract arguments relevant to Client instances.

    :param kwargs: dict, user-provided kwargs
    :return: client arguments dictionary, parameter dictionary
    """
    client_kwargs = {'username': None,
                     'password': None,
                     'access_token': None,
                     'sampler': None,
                     'configuration': 'default',
                     'conf_path': None,
                     'url': 'https://api.qci-prod.com'}
    for key in client_kwargs:
        if key in kwargs:
            # We use `pop` so that items do not get passed to the
            # server if not necessary.
            client_kwargs[key] = kwargs.pop(key)
    return client_kwargs, kwargs


class HTTPClient:

    # default paths to look for the user configuration file
    CONF_PATHS = ('~/.qci.conf', '~/.qci/qci.conf')

    def __init__(self,
                 username: str = None,
                 password: str = None,
                 access_token: str = None,
                 configuration: str = 'default',
                 conf_path: str = None,
                 url: str = 'https://api.qci-prod.com'):
        """Initialize the HTTPClient.

        :param :dict:`options: { url: str, username: str, password: str, access_token: str }`
        """

        self.url = url
        self.username = username
        self.password = password
        self.access_token = access_token
        self.configuration = configuration

        conf = self.get_config(path=conf_path)

        self.conf = conf[self.configuration]
        # fill username, password, access_token with conf file values if the former are None
        self.set_attributes()

        # endpoints
        self.post_endpoint = '/queues/submit'
        self.status_endpoint = '/queues/status?JobID={}'
        self.result_endpoint = '/queues/result?JobID={}'

        self.JobId = None

        # delay in seconds
        self.delay = 1
        self.delay_incr = 1
        self.max_delay = 30

    def __authenticate(self) -> str:
        r"""Try to authenticate against the qbsolv /authorize endpoint"""

        user = {}
        if self.username:
            user['Username'] = self.username

        if self.password:
            user['Password'] = self.password

        if self.access_token:
            user['AccessToken'] = self.access_token

        try:
            res = requests.post(self.url + '/authorize', json=user)
            return res.json()
        except requests.HTTPError as e:
            # TODO: DY, check error
            raise e

    def get_config(self, path=None):
        if path is not None:
            path = os.path.expanduser(path)
            try:
                with open(path, 'r') as fh:
                    conf = yaml.safe_load(fh)
            except FileNotFoundError as e:
                raise e
        else:
            for pth in self.CONF_PATHS:
                # Logic: if the conf file exists and we successfully read it in then
                # we drop through to 'else' and break out of the loop. If 'pth' doesn't exist
                # then continue through path options. If yaml cannot load it, then throw this
                # as a legitimate error.
                try:
                    with open(os.path.expanduser(pth), 'r') as fh:
                        conf = yaml.safe_load(fh)
                except FileNotFoundError:
                    continue
                except yaml.YAMLError as e:
                    raise e
                else:
                    break
        return conf

    def set_attributes(self):
        """
        Setting attributes in this way is a little awkward looking, but it allows the user to override the values in
        the .qci.conf file by passing in args if desired.
        """
        self.username = self.username if self.username is not None else self.conf.get('username')
        self.password = self.password if self.password is not None else self.conf.get('password')
        self.access_token = self.access_token if self.access_token is not None else self.conf.get('access_token')

    def check_status(self, res):
        status = res.json()['Status']
        return status == 'COMPLETE', status == 'ERROR'

    def increment_delay(self):
        """
        Increase the polling delay until we reach 60 seconds between checks. Then check back every 60 seconds.

        :return: None
        """
        if self.delay < self.max_delay:
            self.delay += self.delay_incr
        else:
            self.delay = self.max_delay

    def get_token(self):
        if self.username or self.access_token:
            return self.__authenticate()
        else:
            return {'access_token': ''}

    def create_callback(self, enc):
        encoder_len = enc.len
        # get a starting time and return an instance of the TicToc class
        tic = TicToc()

        def callback(monitor, bar_length=20):
            # mean bytes per second and estimated time remaining
            # tic.toc() takes latest time and subtracts start time
            mean_bps = monitor.bytes_read / tic.toc()
            est_remaining = 'Est. remaining (sec): {}'.format(round((encoder_len - monitor.bytes_read) / mean_bps, 1))
            # Calculate fraction of upload to completed and redo the progress bar
            fractional_progress = round(monitor.bytes_read / encoder_len, 2)
            progress = round(100.0 * fractional_progress, 1)
            tick_len = int(round(bar_length * fractional_progress))
            bar = '=' * tick_len + '-' * (bar_length - tick_len)
            # stdout stays on the same line, and \r moves the cursor back to the start
            sys.stdout.write('\rUpload progress [{}] | {}% | {}'.format(bar, progress, est_remaining))
            sys.stdout.flush()

        return callback

    def post_request(self,
                     data: dict,
                     token: dict,
                     timeout=21600):
        """
        Post a request to the API endpoint at self.post_endpoint.

        :param data: dict, data package
        :param token: dict, contains {'access_token': access_token}
        :param timeout: int, length of time to wait for a job creation response. In seconds.
        :return: None
        """
        with requests.Session() as session:
            enc = encoder.MultipartEncoder(fields=data)
            callback = self.create_callback(enc)
            monitor = encoder.MultipartEncoderMonitor(enc, callback)
            headers = {'Authorization': 'Bearer ' + token['access_token'], "Content-Type": monitor.content_type}
            resp = session.post(self.url + self.post_endpoint, timeout=timeout, headers=headers, data=monitor)

            resp.raise_for_status()

            # Make sure the post confirm that the problem was created (== 201)
            if resp.status_code not in [200, 201]:
                raise requests.HTTPError("Job not successfully created.")

            # Report JobId to the user
            self.JobId = resp.json()['JobID']

            # The job is asynchronous. User can check status using the JobId
            sys.stdout.write('\nJob started with JobID: {}.\n'.format(self.JobId))

    def get_status_api(self, token):
        return requests.get(self.url + self.status_endpoint.format(self.JobId),
                            headers={'Authorization': 'Bearer ' + token['access_token']})

    def report_status_to_user(self, complete, error, delay):
        if complete:
            sys.stdout.write('\nProblem COMPLETE, returning solution.\n')
        elif error:
            sys.stdout.write('\nEncountered ERROR, check response.\n')
        else:
            sys.stdout.write('\nProblem not complete. Checking again in {} seconds...'.format(delay))

    def poll_status(self, token: dict):
        """
        Check status of JobId at increasing intervals. Starts at Default is adding 2 seconds after each check, until
        max_delay is reached (default = 60 sec)

        :param token: dict, {'access_token', access_token}
        :return: None
        """
        COMPLETE = False
        ERROR = False
        while not COMPLETE and not ERROR:
            time.sleep(self.delay)
            res = self.get_status_api(token)
            res.raise_for_status()
            self.increment_delay()
            COMPLETE, ERROR = self.check_status(res)
            self.report_status_to_user(COMPLETE, ERROR, self.delay)

        return COMPLETE, ERROR

    def sendRequest(self, data, large_data=False, timeout=21600) -> Any:
        """
        Send a request to the web service
        :param data: dict, contains 'metadata', 'params', and 'data'
        :param large_data: bool, toggle info message if the data is large.
        :param timeout: int, seconds to wait until giving up on posting to the API # TODO: do we still need this?
        :return: JSON response from server
        """
        if large_data:
            print("Data may take a moment to transfer to the server... ")
        token = self.get_token()
        self.post_request(data, token, timeout)
        complete, error = self.poll_status(token)
        if error:
            # go get the final status so we can send it to the user
            error_res = self.get_status_api(token)
            raise Exception("Failed with Exception {}".format(error_res))
        else:
            # get the completed result, or error if there is one
            res = requests.get(self.url + self.result_endpoint.format(self.JobId),
                               headers={'Authorization': 'Bearer ' + token['access_token']})

            return json.loads(decodeJson(res.json()))


class ClientService(HTTPClient):

    PROBLEM_TYPES = ['qubo', 'constraint_problem', 'graph', 'lagrange_opt']
    SAMPLER_TYPES = ['qbsolv']

    def __init__(self,
                 username: str = None,
                 password: str = None,
                 access_token: str = None,
                 sampler: str = None,
                 configuration: str = 'default',
                 conf_path: str = None,
                 url: str = 'https://api.qci-prod.com'):
        """
        Handles specific data processing for input, as well as communicating with the HTTPClient for authorization and
        sending requests to the API.

        Requires a username and password, or a username and access_token in order to complete authorization

        :param username: str
        :param password: str
        :param access_token: str
        :param sampler: str, which sampler to use. Current list is ['qbsolv'].
        :param configuration: str, configuration name, in case the config file contains multiples
        :param conf_path: str, optional path to local configuration file. By default, HttpClient will check in
            these locations: ('~/.qci.conf', '~/.qci/qci.conf').
        :param url: str, Portal-API server.
        """
        super().__init__(username, password, access_token, configuration, conf_path, url)

        self.sampler = self.conf['sampler'] if sampler is None else sampler
        assert self.sampler in self.SAMPLER_TYPES, ValueError("Sampler must be in [{}]".format(self.SAMPLER_TYPES))

    def _process_problem_type(self, problem_type):
        pass

    def _get_data_type(self, data) -> str:
        if isinstance(data, dict):
            return 'dict'
        elif isinstance(data, np.ndarray):
            return 'dense'
        elif isinstance(data, nx.Graph):
            return 'graph'
        elif isinstance(data, sp.spmatrix):
            return 'sparse'
        else:
            raise TypeError('Unknown data type {}. Must pass dictionary, np.ndarray, '
                            'scipy sparse matrix, or NetworkX Graph'.format(type(data)))

    def get_available_samplers(self):
        """
        Provide a list of the available samplers.

        :return: List of samplers available to users.
        """
        return self.SAMPLER_TYPES

    def process_metadata(self, problem_type, data) -> dict:
        """
        Formulate the metadata dictionary based on input parameters and user input. We need to inform the backend
        services what type of problem we want to solve, the type of data sent, what sampler to use, and (optionally)
        the shape of the data so an encoded matrix can be reconstructed on the server side.

        :param problem_type: str, one of 'qubo', 'constraint_problem', 'lagrange_opt', 'graph'
        :param data:

        :return: dict with {problem_type, data_type, sampler, [shape]}
        """
        assert problem_type in self.PROBLEM_TYPES, \
            ValueError('Problem type must be one of {}'.format(self.PROBLEM_TYPES))
        m = {'problem_type': problem_type,
             'data_type': self._get_data_type(data),
             'sampler': self.sampler}
        if isinstance(data, np.ndarray) or isinstance(data, sp.spmatrix):
            m['shape'] = data.shape
        return encode_data(m)

    def process_params(self, params):
        return encode_data(params)

    @staticmethod
    def process_data(data: Union[dict, np.ndarray, sp.spmatrix, nx.Graph]) -> dict:
        """
        Process the input data by encoding it.

        :param data: dict, np.ndarray, sp.spmatrix, nx.Graph,

        :return: dict with {'data': base64 encoded data object}
        """
        if isinstance(data, dict):
            return {'data': encode_data(data)}
        elif isinstance(data, np.ndarray):
            return encode_ndarray(data)
        elif isinstance(data, sp.spmatrix):
            data = sp.coo_matrix(data, dtype=np.float)
            return encode_data(encode_sparse_array(data))
        elif isinstance(data, nx.Graph):
            data = nx.to_scipy_sparse_matrix(data, nodelist=sorted(set(data.nodes())), dtype=np.float).tocoo()
            return encode_data(encode_sparse_array(data))
        else:
            raise TypeError("Input data must be dictionary, np.ndarray, or scipy.spmatrix")
