#
#    (C) Quantum Computing Inc., 2020.
#
#    THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE of the Copyright Holder. 
#    The contents of this file may not be disclosed to third parties, copied 
#    or duplicated in any form, in whole or in part, without the prior written
#    permission of the Copyright Holder.
#
from typing import Union
import networkx as nx
from scipy.sparse import spmatrix
import numpy as np

from .client import ClientService, get_client_kwargs
from .response_client import ClientGraphResponse

__all__ = ['minimum_clique_cover', 'community_detection', 'partition']


class QNXClient(ClientService):

    def __init__(self,
                 username: str = None,
                 password: str = None,
                 access_token: str = None,
                 url: str = 'https://api.qci-prod.com',
                 configuration: str = 'default',
                 conf_path: str = None,
                 sampler: str = 'qbsolv'):
        """

        :param username: str
        :param password: str
        :param access_token: str
        :param url: str
        :param sampler: str, string representation of the ampler to invoke. Default = 'qbsolv'.
        """
        super().__init__(username, password, access_token, sampler, configuration, conf_path, url)

    def solve_graph_problem(self, G: Union[nx.Graph, np.ndarray, spmatrix], graph_algo: str, **kwargs):
        """
        This function is typically invoked through one of the higher-level interfaces defined below.

        :param G: NetworkX Graph, np.ndarray or  sp.spmatrix.
        :param graph_algo: str, name of graph algorithm.
        :param kwargs: dict, keyword args for the specific algorithm. Eg.,
            k, alpha, beta
        :return: ClientGraphResponse
        """
        if isinstance(G, nx.Graph):
            G = nx.to_scipy_sparse_matrix(G, nodelist=sorted(set(G.nodes())), dtype=np.float).tocoo()
        metadata = self.process_metadata('graph', G)
        params_elements = {'graph_algo': graph_algo}
        params_elements.update(kwargs)
        params = self.process_params(params_elements)
        data_obj = self.process_data(G)
        message = {'data_obj': ('data_obj.txt', data_obj, 'text/plain'),
                   'metadata': ('metadata.txt', metadata, 'text/plain'),
                   'params': ('params.txt', params, 'text/plain')}
        res = self.sendRequest(message)
        return ClientGraphResponse.from_dict_response(res)

    def minimum_clique_cover(self, G: Union[nx.Graph, np.ndarray, spmatrix], chromatic_lb: int = None, chromatic_ub: int = None, **kwargs):
        """
        Partition vertices in an undirected graph into a cover of cliques.

        :param G: NetworkX Graph, np.ndarray or  sp.spmatrix. Undirected, with possible edge weights.
        :param chromatic_lb: int, optional. A lower bound on the chromatic number.
        :param chromatic_ub: int, optional. A upper bound on the chromatic number.
        :param kwargs: Parameters to pass to the algorithm and solver.
            Clique Cover takes 'alpha', a Lagrange multiplier.
            For sampler parameters, see documentation in the sampler module
        :return: ClientGraphResponse
        """
        graph_algo = 'clique_cover'
        kwargs.update({'chromatic_lb': chromatic_lb})
        kwargs.update({'chromatic_ub': chromatic_ub})
        return self.solve_graph_problem(G, graph_algo, **kwargs)

    def community_detection(self, G: Union[nx.Graph, np.ndarray, spmatrix], k: int = 2, **kwargs):
        """
        Partition a graph into k disjoint community of nodes, using community detection.

        :param G: NetworkX Graph, np.ndarray or  sp.spmatrix. Undirected, with possible edge weights.
        :param k: int, number of communities
        :param kwargs: Parameters to pass to the algorithm and solver.
            Community Detection takes 'alpha', the importance of constraining each node to a single community vs modularity.
            For sampler parameters, see documentation in the sampler moduleß
        :return: ClientGraphResponse
        """
        graph_algo = 'community_detection'
        kwargs.update({'k': k})
        return self.solve_graph_problem(G, graph_algo, **kwargs)

    def partition(self, G: Union[nx.Graph, np.ndarray, spmatrix], k: int = 2, **kwargs):
        """
        Partition a graph into k disjoint collections of nodes.

        :param G: NetworkX Graph, np.ndarray or  sp.spmatrix. Undirected, with possible edge weights.
        :param k: int, number of partitions, default = 2.
        :param kwargs: Parameters to pass to the algorithm and solver.
            Partition takes 'alpha' and 'beta', the importance of minimizing edge cuts and
            the importance of a balanced partition, respectively.
            For sampler parameters, see documentation in the sampler moduleß
        :return: ClientGraphResponse
        """
        graph_algo = 'partition'
        kwargs.update({'k': k})
        return self.solve_graph_problem(G, graph_algo, **kwargs)


def minimum_clique_cover(G: Union[nx.Graph, np.ndarray, spmatrix], chromatic_lb: int = None, chromatic_ub: int = None, **kwargs):
    """
    Partition vertices in a graph into a clique cover with minimum cardinality.

    :param G: NetworkX Graph, np.ndarray or  sp.spmatrix. Undirected, with possible edge weights. Requirement: For NetworkX Graph, integer nodes (0-based indexing) required (i.e. - For 3-nodes graph, G.nodes = NodeView((0, 1, 2)))
    :param chromatic_lb: int, optional. A lower bound on the chromatic number. Default = None.
    :param chromatic_ub: int, optional. A upper bound on the chromatic number. Default = None.
    If *chromatic_up* is None, then is estimated from

        :math:`\\min \\left( \\lambda_{\\text{max}}(A), \\left\\lceil {1 + \\sqrt{1+ 8|E|}}\\right\\rceil / 2 \\right)`,

    where :math:`A` is the adjacency matrix of :math:`G`, and :math:`\\lambda_{\\text{max}}` indicates the maximum eigenvalue.

    Similarly, if *chromatic_lb* is None, it will be computed as the *clique covering number*, see `Clique covering number <https://mathworld.wolfram.com/CliqueCoveringNumber.html>`_.

    :param kwargs: Parameters to pass to the algorithm and solver.
        For sampler parameters, see documentation in the sampler module.

        * *alpha* float, optional. A factor controlling the importance of the orthogonality of the cliques to each other (see [DA]_ Alg 15).  Empirical guidance on good values for *alpha* is being explored.

    :return: ClientGraphResponse
    """

    client_kwargs, param_kwargs = get_client_kwargs(**kwargs)
    client = QNXClient(**client_kwargs)
    return client.minimum_clique_cover(G, chromatic_lb=chromatic_lb, chromatic_ub=chromatic_ub, **param_kwargs)


def community_detection(G: Union[nx.Graph, np.ndarray, spmatrix], k: int = 2, **kwargs):
    """
    Partition a graph into k disjoint communities of nodes, maximizing modularity.

    :param G: NetworkX Graph, np.ndarray or  sp.spmatrix. Undirected, with possible edge weights. Requirement: For NetworkX Graph, integer nodes (0-based indexing) required (i.e. - For 3-nodes graph, G.nodes = NodeView((0, 1, 2)))
    :param k: int, number of communities, default = 2.
    :param kwargs: Parameters to pass to the algorithm and solver.
        For sampler parameters, see documentation in the sampler module

        * *alpha* float, optional. A factor controlling the importance of constraining each node to a single community, in trade for the importance of modularity.

    :return: ClientGraphResponse
    """

    client_kwargs, param_kwargs = get_client_kwargs(**kwargs)
    client = QNXClient(**client_kwargs)
    return client.community_detection(G, k=k, **param_kwargs)


def partition(G: Union[nx.Graph, np.ndarray, spmatrix], k: int = 2, **kwargs):
    """
    Partitions a graph into k disjoint collections of nodes, minimizing the number of inter-partition edges.

    :param G: NetworkX Graph, np.ndarray or  sp.spmatrix. Undirected, with possible edge weights. Requirement: For NetworkX Graph, integer nodes (0-based indexing) required (i.e. - For 3-nodes graph, G.nodes = NodeView((0, 1, 2)))
    :param k: int, number of partitions, default = 2.
    :param kwargs: Parameters to pass to the algorithm and solver.
        For sampler parameters, see documentation in the sampler module

        * *alpha* float, optional. A factor controlling the importance of a balanced partition.
        * *beta* float, optional. A factor controlling the importance of minimizing edge cuts.

    :return: ClientGraphResponse
    """

    client_kwargs, param_kwargs = get_client_kwargs(**kwargs)
    client = QNXClient(**client_kwargs)
    return client.partition(G, k=k, **param_kwargs)
