#
#    (C) Quantum Computing Inc., 2020.
#
#    THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE of the Copyright Holder. 
#    The contents of this file may not be disclosed to third parties, copied 
#    or duplicated in any form, in whole or in part, without the prior written
#    permission of the Copyright Holder.
#
from typing import Union
import numpy as np
import scipy.sparse as sp

from .client import ClientService, get_client_kwargs
from .response_client import ClientResponse
from .utils import check_qubo, check_constraints, check_qubo_dict, convert_dict_to_sparse_qubo

__all__ = ['sample_qubo', 'sample_constraint_problem', 'sample_lagrange_optimization']


class QuoirClient(ClientService):

    def __init__(self,
                 username: str = None,
                 password: str = None,
                 access_token: str = None,
                 sampler: str = 'qbsolv',
                 configuration: str = 'default',
                 conf_path: str = None,
                 url: str = 'https://api.qci-prod.com'):
        """

        :param username: str
        :param password: str
        :param access_token: str
        :param url: str
        :param sampler: str, string representation of the ampler to invoke. Default = 'qbsolv'.
        """
        super().__init__(username, password, access_token, sampler, configuration, conf_path, url)

    def _check_qubo(self, qubo: Union[dict, np.ndarray, sp.spmatrix]):
        """
        Make sure that the QUBO is properly constructed. Check that it is square and symmetric.

        :param qubo: dict, np.ndarray or sp.spmatrix, User-input QUBO

        Raises AssertionError if QUBO is not square or symmetric

        :return: None
        """
        if isinstance(qubo, dict):
            # raises AssertionError if anything is amiss
            check_qubo_dict(qubo)
        else:
            # raises AssertionError if anything is amiss
            check_qubo(qubo)

    def _check_constraints(self, objective, constraints):
        """

        :param objective: np.ndarray or sp.spmatrix, User-input objective function
        :param constraints: np.ndarray or sp.spmatrix, User-input constraint matrix

        Raises AssertionError if Objective is not square or symmetric, or

        :return: None
        """
        self._check_qubo(objective)
        check_constraints(objective, constraints)

    def sample_qubo(self,
                    qubo: Union[dict, np.ndarray, sp.spmatrix],
                    **params):
        """
        Given a Quadratic Unconstrained Binary Optimization model (QUBO), return samples of optimal or near optimal
        solutions.

        :param qubo: dict, np.ndarray, or sp.spmatrix, n x n QUBO matrix
        :param sampler: str

        :param params:
            Parameters and arguments for the sampler and solver algorithm. Different samplers permit different arguments.

        :return: ClientResponse
        """
        # make sure the input is correct before
        self._check_qubo(qubo)
        if isinstance(qubo, dict):
            qubo = convert_dict_to_sparse_qubo(qubo)
        metadata = self.process_metadata('qubo', qubo)
        params = self.process_params(params)
        data_obj = self.process_data(qubo)
        message = {'data_obj': ('data_obj.txt', data_obj, 'text/plain'),
                   'metadata': ('metadata.txt', metadata, 'text/plain'),
                   'params': ('params.txt', params, 'text/plain')}
        res = self.sendRequest(message)
        return ClientResponse.from_dict_response(res)

    def sample_constraint_problem(self,
                                  objective: Union[np.ndarray, sp.spmatrix],
                                  constraints: Union[np.ndarray, sp.spmatrix],
                                  alpha: float = 1.0,
                                  **params) -> ClientResponse:
        """
        Given an objective function and linear constraint matrix, find optimal or near optimal solutions to the
        associated QUBO,

        :math:`Q = B + alpha * C.T * C,`

        where :math:`B` is the :math:`n x n` objective function matrix, and :math:`C` is the linear constraint matrix.
        The dimension of :math:`C` is :math:`m x n+1`, where the constraint vector is appended to the final column of
        :math:`C`.

        :param objective: np.ndarray or sp.spmatrix, n x n cost function matrix
        :param constraints: np.ndarray or sp.spmatrix, m x n or m x n+1 constraint functions matrix, where
            m is the number of constraints and n is the dimension of the solutions space.
        :param alpha: float, Lagrange multiplier to scale constraints when computing QUBO,
            Q = objective + alpha * constraints
            Default = 1.0
        :param params: Parameters for the provided sampler. For allowed arguments see documentation
            on in the sampler module.

        :return: ClientResponse
        """
        assert alpha > 0, "Lagrangian multiplier 'alpha' must be positive."
        self._check_constraints(objective, constraints)
        params.update({'alpha': alpha})
        metadata = self.process_metadata('constraint_problem', constraints)
        params = self.process_params(params)
        data_obj = self.process_data(objective)
        data_con = self.process_data(constraints)
        message = {'data_obj': ('data_obj.txt', data_obj, 'text/plain'),
                   'data_con': ('data_con.txt', data_con, 'text/plain'),
                   'metadata': ('metadata.txt', metadata, 'text/plain'),
                   'params': ('params.txt', params, 'text/plain')}
        res = self.sendRequest(message)
        return ClientResponse.from_dict_response(res)

    def sample_lagrange_optimization(self,
                                     objective: Union[np.ndarray, sp.spmatrix],
                                     constraints: Union[np.ndarray, sp.spmatrix],
                                     **kwargs):
        """
        An iterative solver that uses a modified gradient descent algorithm to find the optimal
        Lagrange multiplier, alpha, for the Qubo equation,

            Q = objective + alpha * constraints

        :param objective: np.ndarray or sp.spmatrix, :math:`n x n` cost function matrix
        :param constraints: np.ndarray or sp.spmatrix, :math:`m x n+1` constraint functions matrix, where
            :math:`m` is the number of constraints and :math:`n` is the dimension of the solutions space.
        :param kwargs: Parameters for the provided sampler and optimization.
            For allowed arguments see documentation on in the sampler module.
            For Lagrange optimization, **max_depth**, int,  corresponds to the number of iterations allowed for the
            optimization loop. Default = 5.

        :return: ClientResponse
        """
        self._check_constraints(objective, constraints)
        if kwargs.get('max_depth') is None:
            kwargs['max_depth'] = 5
        # if the user input 'max_depth', max sure it's nonnegative
        assert kwargs['max_depth'] >= 0, "Parameter 'max_depth' must be nonnegative."
        metadata = self.process_metadata('lagrange_opt', constraints)
        params = self.process_params(kwargs)
        data_obj = self.process_data(objective)
        data_con = self.process_data(constraints)
        message = {'data_obj': ('data_obj.txt', data_obj, 'text/plain'),
                   'data_con': ('data_con.txt', data_con, 'text/plain'),
                   'metadata': ('metadata.txt', metadata, 'text/plain'),
                   'params': ('params.txt', params, 'text/plain')}
        res = self.sendRequest(message)
        return ClientResponse.from_dict_response(res)


def sample_qubo(qubo: Union[dict, np.ndarray, sp.spmatrix], **kwargs) -> ClientResponse:
    """
    Return samples of (near) optimal solutions for a Quadratic Unconstrained Binary Optimization (QUBO) problem.

    :param qubo: dict, np.ndarray, or sp.spmatrix, :math:`n \\times n` QUBO matrix.
        Matrices must be symmetric. Dictionary input can represent the upper triangular portion of a matrix
        (including the diagonal) or the full matrix.

    :param kwargs:
        Parameters and arguments for the sampler and solver algorithm.
        Optional login information for the cloud client.

    :return: :meth:`mukai.response_client.ClientResponse`

    :Example:

    QUBO array input using a Numpy array (see [GK]_):

    >>> import numpy as np
    >>> from mukai import quoir
    >>> # 'qubo' is a symmetric version of a quadratic unconstrained binary optimization (QUBO) problem
    >>> qubo = np.array([[-6, 10, 0, 0], [10, -8, 2, 8], [0, 2, -3, 4], [0, 8, 4, -5]])
    >>> # sample_qubo requires a symmetric matrix
    >>> res = quoir.sample_qubo(qubo, num_solutions=5)
    >>> res
    >>>  # Job started with JobID: 5ef37a0d08b481fef4af1d1f.
    >>>     # Problem COMPLETE, returning solution.
    >>>     # ClientResponse
    >>>     # samples
    >>>     # [[1. 0. 0. 1.]
    >>>     #  [1. 0. 1. 0.]
    >>>     #  ...
    >>>     #  [0. 0. 1. 1.]]
    >>>     # energies
    >>>     # [-11.  -9.  -8.  -7.  -6.  -6.  -5.  -3.   0.   0.]
    >>>     # counts
    >>>     # [1 1 1 1 1 1 1 1 1 1]
    """
    client_kwargs, param_kwargs = get_client_kwargs(**kwargs)
    client = QuoirClient(**client_kwargs)
    return client.sample_qubo(qubo, **param_kwargs)


def sample_constraint_problem(objective: Union[np.ndarray, sp.spmatrix],
                              constraints: Union[np.ndarray, sp.spmatrix],
                              alpha: float = 1.0,
                              **kwargs) -> ClientResponse:
    """
    Return samples of a constraint problem with an objective function and linear equality constraints.

    The objective function is defined by a :math:`n \\times n` matrix; the constraint matrix is :math:`m \\times n+1`,
    which is with a constraint vector in the final column.

    Example:

    >>> objective = np.array([[-6, 10, 0, 0], [10, -8, 2, 8], [0, 2, -3, 4], [0, 8, 4, -5]])
    >>> constraints = np.array([[1, 1, 1, 1, -1], [0, 1, 1, 1, -1]])

      where *objective* is an :math:`n \\times n` cost function and *constraints* is an :math:`m \\times (n+1)` matrix composed of a
      system of linear constraints. One can think of constraints as an *augmented* system of linear constraints. In the
      example above this would be

    >>> A = constraints[:, :-1]

        and the final column

    >>> b = np.array([1, 1.])

        giving  *constraints* = [A | -b].

    If the variables are labeled by the column indices, then the rows of the augmented constraint matrix state that

            :math:`x_0 + x_1 + x_2 + x_3 - 1 == 0`

            :math:`x_1 + x_2 + x_3 - 1 == 0`

    To compute the QUBO we combine the objective and constraints.

    :param objective: np.ndarray or sp.spmatrix, :math:`n \\times n` symmetric cost function matrix.
    :param constraints: np.ndarray or sp.spmatrix, :math:`m \\times n+1` constraint matrix, where
       :math:`m` is the number of constraints and :math:`n` is the dimension of the solution space.
    :param alpha: float, optional. Lagrange multiplier to scale constraints when computing QUBO,
       *Q = objective + alpha * constraints*. Default = 1.0.

    :param kwargs:
        Parameters for the provided sampler. For supported arguments see documentation
        on the sampler module.
        * **sampler** -- str, the sampler to call. Default = *qbsolv*.

    :return: :meth:`mukai.response_client.ClientResponse`

    :Example:

    >>> import numpy as np
    >>> from mukai import quoir
    >>> objective = np.array([[-6, 10, 0, 0], [10, -8, 2, 8], [0, 2, -3, 4], [0, 8, 4, -5]])
    >>> constraints = np.array([[1, 1, 1, 1, -1], [0, 1, 1, 1, -1]])
    >>> alpha = 2
    >>> res = quoir.sample_constraint_problem(objective, constraints, alpha=alpha, num_solutions=5)
    """
    client_kwargs, param_kwargs = get_client_kwargs(**kwargs)
    client = QuoirClient(**client_kwargs)
    return client.sample_constraint_problem(objective, constraints, alpha, **param_kwargs)


def sample_lagrange_optimization(objective: Union[np.ndarray, sp.spmatrix],
                                 constraints: Union[np.ndarray, sp.spmatrix],
                                 **kwargs) -> ClientResponse:
    """
    Return samples of an optimal solution to a constraint optimization problem with an objective function
    and a system of linear constraints.

    Consider the problem defined by an objective function represented by an :math:`n \\times n` matrix, and an :math:`m \\times n+1`
    constraint matrix. See :meth:`sample_constraint_problem` for details on the *objective* and *constraints* matrices.
    The Lagrange multiplier is designed to balance the scales of the objective function and the system of constraints.

    Ideally, the Lagrangian, :math:`\\alpha`, reflects the importance of the constraints, and can be a vector.
    Lagrange Optimization is an iterative approach for choosing the ideal :math:`\\alpha` such that the constraints are satisfied, allowing the
    sampler to focus on minimizing the energy of the objective function.

    :Example:

    >>> objective = np.array([[-6, 10, 0, 0], [10, -8, 2, 8], [0, 2, -3, 4], [0, 8, 4, -5]])
    >>> constraints = np.array([[1, 0, 1, 1, -1], [1, 1, 0, 1, -1]])

    To compute the QUBO we combine the objective and constraints.

    Lagrange Optimization runs the chosen sampler up to `max_depth` times to determine the :math:`\\alpha` that
    minimizes *Q* while simultaneously satisfying constraints, eg., :math:`Av_0 == 0`, where :math:`v_0` is the
    minimum energy sample returned by the sampler.

    Lagrange optimization is more involved, and returns an additional non-empty `properties` attribute with the
    ClientResponse object. In `properties['alpha']`, the :math:`m`-dimensional vectors record the updates to `alpha`,
    which are performed on a
    per-constraint basis, where :math:`m` is the dimension of the row space of the constraint matrix, :math:`A`.
    The  `properties['constraint_evals']` values record the result of :math:`Av_0` for each iteration.

    :param objective: np.ndarray or sp.spmatrix, :math:`n \\times n` symmetric cost function matrix.
    :param constraints: np.ndarray or sp.spmatrix, :math:`m \\times n+1` constraint matrix, where
       :math:`m` is the number of constraints and :math:`n` is the dimension of the solution space.
    :param kwargs: Parameters for the provided sampler. For supported arguments see documentation
        on the sampler module.

    :Keyword Arguments:

        * **max_depth** -- int, maximum number of times to update *alpha* in a search for a value that will satisfy
          constraints as well as minimize energy.

          Note: Indexing for depth (iterations) is 0-based. The `sample_lagrange_optimization`
          function performs at minimum a single (0th) pass. This is consistent with the concept of depth, in that
          a surface-level analysis is `depth=0`, and further exploration, `depth > 1`, implies digging deeper. Furthermore,
          this implies that for `depth=k`, the number of vectors in the `alpha` field of `ClientResponse.properties` will
          be `k+1`.

        * **sampler** -- str, the sampler to call. Default = *qbsolv*.

    :return: :meth:`mukai.response_client.ClientResponse`

    :Example:

    >>> import numpy as np
    >>> from mukai import quoir
    >>> objective = np.array([[ 3.4 ,  1.1 , -1.7 , -1.6 ], [ 1.1 ,  4.5 ,  2.25,  2.25],  [-1.7 ,  2.25,  2.1 ,  0.5 ],  [-1.6 ,  2.25,  0.5 , -2.4 ]])
    >>> constraints = np.array([[1, 0, 1, 1, -1], [1, 1, 0, 1, -1]])
    >>> res = quoir.sample_lagrange_optimization(objective, constraints, max_depth=20, num_solutions=5)
    """
    client_kwargs, param_kwargs = get_client_kwargs(**kwargs)
    client = QuoirClient(**client_kwargs)
    return client.sample_lagrange_optimization(objective, constraints, **param_kwargs)
