#
#    (C) Quantum Computing Inc., 2020.
#
#    THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE of the Copyright Holder. 
#    The contents of this file may not be disclosed to third parties, copied 
#    or duplicated in any form, in whole or in part, without the prior written
#    permission of the Copyright Holder.
#
from abc import ABC, abstractmethod
from typing import Dict, Any, Union
import numpy as np
import scipy.sparse as sp
import pandas as pd
import networkx as nx


class ABCResponse(ABC):
    @abstractmethod
    def __init__(self):
        super().__init__()

    @abstractmethod
    def _get_name(self):
        pass

    @abstractmethod
    def get_num_samples(self):
        pass

    @abstractmethod
    def get_samples(self):
        pass

    @abstractmethod
    def to_json_response(self):
        pass


class ClientResponse(ABCResponse):

    def __init__(self, samples: Any, energies: Any, counts: Any,
                 time: dict = None, properties: dict = None):
        """
        ClientResponse is the base class containing method for access and analyzing
        the samples, energies, and counts returned by QuoirClient.

        Args:
            samples: iterable, such as list[dict sample]
            energies: iterable of float energies
            counts: iterable of int counts
            time: dict (optional), table of timing information related to setup and execution of the sampler
            properties: dict (optional), Container for additional information.
                Must be json-serializable if planning to use to_json_response later.
        """
        super().__init__()
        self.samples = samples
        self.energies = energies
        self.counts = counts
        self.time = time
        self.properties = properties

    @classmethod
    def from_dict_response(cls, resp):
        # num samples x num dimensions
        samples = np.zeros((len(resp['samples']), len(resp['samples'][0])))
        for i, sample in enumerate(resp['samples']):
            # sort the indices -- we assume integer indices for now
            sample = sorted(sample.items(), key=lambda u: int(u[0]))
            samples[i] = [sample[i][1] for i in range(len(sample))] 
        return cls(samples,
                   np.asarray(resp['energies']),
                   np.asarray(resp['counts']),
                   resp.get('time', {}),
                   resp.get('properties', {}))
   
    @classmethod
    def filter_feasible_solutions(cls, resp, constraints: Union[np.ndarray, sp.spmatrix]):
        # assumes samples[i] np.array 
        # Initialize empty container for samples that meet constraints
        meet_constraints_samples = []
        meet_constraints_energies = []
        meet_constraints_counts = []
        b = - constraints[:, -1]
        A = constraints[:, :-1]
        for i, sample in enumerate(resp.samples):
            if np.all(np.matmul(A, resp.samples[i].T)==b):
                meet_constraints_samples.append(resp.samples[i])
                meet_constraints_energies.append(resp.energies[i])
                meet_constraints_counts.append(resp.counts[i])
        return cls(np.asarray(meet_constraints_samples), np.asarray(meet_constraints_energies), np.asarray(meet_constraints_counts))

    def _get_name(self):
        return "ClientResponse"

    def __repr__(self) -> str:
        repr = [self._get_name()]
        for name, data in zip(['samples', 'energies', 'counts'],
                              [self.samples, self.energies, self.counts]):
            repr.append('\n{}\n{}'.format(name, data[:10]))
        return ''.join(repr)

    def get_num_samples(self) -> int:
        return len(self.samples)

    def get_samples(self) -> Any:
        """
        Returns:
            array-like, samples
        """
        return self.samples

    def to_pandas_response(self) -> Any:
        """
        Construct a Pandas DataFrame using samples, energies, and counts attributes

        :return: pandas DataFrame
        """
        d = {"samples": list(self.samples), "energies": self.energies, "counts": self.counts}
        return pd.DataFrame(data=d)

    def to_json_response(self) -> Dict:
        """
        :return: dict
        """
        return {'samples': self.samples,
                'energies': self.energies,
                'counts': self.counts,
                'time': self.time,
                'properties': self.properties}


class ClientGraphResponse(ClientResponse):

    def __init__(self, samples: Any, encoded_samples: Any, energies: Any, counts: Any, time: dict = None):
        """
        ClientGraphResponse is the base class containing method for access and analyzing
        the samples, energies, and counts returned by QCI-Networkx-QuOIR Client.

        Args:
            samples: iterable, such as list[dict sample]
            encoded_samples: iterable such as list[list encoded samples]
            energies: iterable of float energies
            counts: iterable of int counts
            time: dict (optional), table of timing information related to setup and execution of the sampler
        """
        super().__init__(samples, energies, counts, time)
        self.encoded_samples = encoded_samples

    @classmethod
    def from_dict_response(cls, resp):
        return cls(resp['samples'],
                   resp['encoded_samples'],
                   np.asarray(resp['energies']),
                   np.asarray(resp['counts']),
                   resp['time'])

    def to_json_response(self) -> Dict:
        """
        Returns:
            dict
        """
        return {'samples': self.samples,
                'encoded_samples': self.encoded_samples,
                'energies': self.energies,
                'counts': self.counts,
                'time':self.time}

    def from_dict_to_array(self, data):
        return [i for i in data.values()]

    def _get_name(self):
        return "ClientGraphResponse"

    def map_sample_to_graph(self, G: Union[nx.Graph, np.ndarray, sp.spmatrix], sample: dict):
        """
        Map input sample to node attributes of input graph, G.

        Args:
            :param G: networkx graph, numpy ndarray or scipy spmatrix. Undirected, with possible edge weights.
            :param sample: dict, nodes and their respective label values
        """
        # TODO: Current implementation causes data mutation of input data. Need to edit so that it doesn't cause mutation on input data
        # Convert to networkx graph structure if input graph is in dense or sparse matrix representation
        G_copy = G # To avoid mutation of input parameter
        if isinstance(G_copy, np.ndarray):
            G_copy = nx.from_numpy_array(G_copy)
        elif isinstance(G_copy, sp.spmatrix):
            G_copy = nx.from_scipy_sparse_matrix(G_copy)
        # Set attribute label of graph nodes
        nx.set_node_attributes(G_copy, sample, 'label')
        return G_copy
