#
#    (C) Quantum Computing Inc., 2020.
#
#    THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE of the Copyright Holder. 
#    The contents of this file may not be disclosed to third parties, copied 
#    or duplicated in any form, in whole or in part, without the prior written
#    permission of the Copyright Holder.
#
import time
from typing import Union
import numpy as np
import scipy.sparse as sp
from itertools import chain


class TicToc(object):

    def __init__(self):
        self.start = time.time()

    @classmethod
    def tic(cls):
        return cls()

    def toc(self):
        return time.time() - self.start


def map_nodes(qubo: dict) -> dict:
    """
    Maps unique tuples of length k representing nodes in a graph to integers.

    Args:
        qubo: QUBO in dictionary form with tuples as keys

    Returns:
        dict[tuple, int], mapping from key to unique integer
    """
    nodes = sorted(set(chain(*qubo.keys())))
    return {u: i for i, u in enumerate(nodes)}


def construct_qubo_int_keys(qubo: dict):
    """
    Construct a mapping from keys of a QUBO to integer keys. Currently, options are tuple or string source keys.

    Args:
        qubo: dict, QUBO in sparse dict form.

    Returns:
        dict of QUBO with integer nodes/keys, with associated mapping from original nodes to integer nodes
    """
    node_map = map_nodes(qubo)
    Qint = {}
    for (u, v), val in qubo.items():
        Qint[node_map[u], node_map[v]] = val
    return {'qubo': Qint, 'node_map': _invert_mapping(node_map)}


def _invert_mapping(mapping: dict) -> dict:
    return {v:k for k,v in mapping.items()}


def map_solutions(solns, mapping):
    return [map_solution(soln, mapping) for soln in solns]


def map_solution(soln, mapping):
    """
    Invert a dictionary `mapping`.

    Args:
        soln: dict[int, val], logical variable or graph node and sampled value (float). Assumed that
            keys are ints due to limitation of QBSolv.
        mapping: dict[int, (any hashable)], mapping ints to non-integer variable indices.

        The `mapping` must have the same values as the keys in `soln`, so that when `mapping`
        is inverted the lookup is well-defined for 'inverted mapping'.

    Returns:
        dict[mapping[int], val]
    """
    return {mapping[i]: val for i, val in soln.items()}


def check_int_nodes(qubo: dict) -> bool:
    idx = set(chain(*qubo.keys()))
    return all(map(lambda m: isinstance(m, int), idx))


def is_upper_tri(qubo: dict) -> bool:
    idx = sorted(set(qubo.keys()))
    return all(map(lambda m: m[0] <= m[1], idx))


def convert_dict_to_sparse_qubo(qubo: dict) -> sp.spmatrix:
    """
    Convert dictionary to sparse matrix format.

    :param qubo: dict, QUBO

    :return: COO matrix
    """
    upper_tri = False
    # we need to divide the influence of each off-diagnonal term.
    if is_upper_tri(qubo):
        upper_tri = True
        get_val = lambda row, col, v: v if row == col else v / 2.
    else:
        get_val = lambda row, col, v: v

    n = max(set(chain(*qubo.keys())))
    mat = sp.dok_matrix((n + 1, n + 1), dtype=np.float)
    for i, j in qubo.keys():
        if not upper_tri:
            assert qubo[i,j] == qubo[j,i]
        mat[i, j] = mat[j, i] = get_val(i, j, qubo[i, j])
    return mat.tocoo()


def check_qubo_dict(qubo: dict):
    # check keys are in [0, 1, ..., n-1]
    assert check_int_nodes(qubo), "QUBO keys must be integers in [0, 1, ..., n-1]."


def check_qubo(qubo: Union[np.ndarray, sp.spmatrix], size_limit=10000):
    assert (isinstance(qubo, np.ndarray) or isinstance(qubo, sp.spmatrix)), \
        "QUBO must be of type ndarray or sparse matrix."
    assert qubo.ndim == 2, "QUBO or Objective must be a 2-dimensional array."
    # check QUBO problems in array form are symmetric and square
    assert qubo.shape[0] == qubo.shape[1], "Input QUBO or Objective must be square. QUBO.shape = {}".format(qubo.shape)
    # check QUBO size is less than or equal to 10k (Mukai 2.0)
    assert qubo.shape[0] <= size_limit, "QUBO must be less than or equal to {} x {}. The input QUBO is QUBO.shape = {}".format(size_limit, size_limit, qubo.shape)
    # these only get called if qubo is square
    if isinstance(qubo, np.ndarray):
        assert check_symmetric_ndarray(qubo), "QUBO or Objective is not symmetric."
    else:
        assert check_symmetric_sparse(qubo), "QUBO or Objective is not symmetric."


def check_constraints(objective: Union[np.ndarray, sp.spmatrix], constraints: Union[np.ndarray, sp.spmatrix]):
    # Make sure the data type is correct
    assert (isinstance(constraints, np.ndarray) or isinstance(constraints, sp.spmatrix)), \
        "Constraint matrix must be of type ndarray or sparse matrix."
    # Now make sure both objective and constraint are the instance type. This check should suffice.
    assert isinstance(objective, constraints.__class__)
    assert constraints.shape[constraints.ndim-1] == objective.shape[0] + 1, \
        "Column dimension of constraint matrix must equal objective.shape[1] + 1. " \
        "Input: constraint.shape = {}, objective.shape = {}".format(constraints.shape, objective.shape)


def check_symmetric_ndarray(a, rtol=1e-06, atol=1e-09):
    """
    See:
    https://stackoverflow.com/questions/48798893/error-in-checking-symmetric-sparse-matrix
    """
    return np.allclose(a, a.T, rtol=rtol, atol=atol)


def check_symmetric_sparse(a, tol=1e-10):
    """
    See the most excellent post here:
    https://stackoverflow.com/questions/48798893/error-in-checking-symmetric-sparse-matrix
    """
    return (abs(a - a.T) > tol).nnz == 0


def loadQuboFile(filename) -> dict:
    qubo = {}
    with open(filename, 'r') as handle:
        for line in handle.readlines():
            line = line.strip()
            if len(line) == 0 or line[0] == 'c':
                continue
            if line[0] == 'p':
                _, _, _, n_variables, _, _ = line.split()
                qubo = {(a, a): 0 for a in range(int(n_variables))}
                continue
            # TODO: we should check that 'p' line has been found first
            a, b, bias = line.split()
            qubo[int(a), int(b)] = qubo.get((int(a), int(b)), 0) + float(bias)
    return qubo